package voicebot;

import java.util.LinkedHashMap;
import java.util.Map;

public abstract class AbstractVoiceBotCommand {

	private final String args;
	private final String info;
	private final Map<String, String> help;
	private final String keys;
	private String[] notes;
	private boolean adminOnly;

	public AbstractVoiceBotCommand(String args, String info, String... help) {
		this.args = args;
		this.info = info;
		this.help = new LinkedHashMap<String, String>();
		notes = new String[0];

		String k = "";
		for (int i = 0; i < help.length; i += 2) {

			if (i + 1 < help.length) {
				this.help.put(help[i], help[i + 1]);
				k += help[i];
			}
		}
		keys = k;
	}

	protected void setNotes(String... notes) {

		if (notes != null) {
			this.notes = notes;
		}
	}

	protected void setAdminOnly() {
		adminOnly = true;
	}

	public boolean isAdminOnly() {
		return adminOnly;
	}

	public String name() {
		return getClass().getSimpleName().toLowerCase();
	}

	public String info() {
		return info;
	}

	protected void sendError(VoiceBot bot, String username, VoiceBotResponses response, String... values) {

		bot.sendNotice(username, response, values);
		throw new SelvvirCommandTerminated();
	}

	protected String truncate(String url) {

		if (url.length() > 31) {
			return url.substring(0, 20) + "…" + url.substring(url.length() - 10);
		}
		return url;
	}

	public void run(VoiceBot bot, String flags, String channel, String sender, String target, String text) {

		if (flags.contains("?")) {
			String pargs = (args.isEmpty() ? "" : args + " ");

			if (help.isEmpty()) {
				bot.sendNotice(target, VoiceBotResponses.COMMAND_SYNTAX_NO_FLAGS, name(), pargs, info);
			}
			else {
				bot.sendNotice(target, VoiceBotResponses.COMMAND_SYNTAX, name(), keys, pargs, info);
				bot.sendNotice(target, VoiceBotResponses.COMMAND_FLAGS);
			}

			for (String key : help.keySet()) {
				bot.sendNotice(target, VoiceBotResponses.COMMAND_FLAG, key, help.get(key));
			}

			for (String note : notes) {
				bot.sendNotice(target, VoiceBotResponses.COMMAND_NOTE, note);
			}
		}
		else {
			runf(bot, flags, channel, sender, target, text);
		}
	}

	public abstract void runf(VoiceBot bot, String flags, String channel, String sender, String target, String text);

	/*
	 * The exit early class.
	 */
	public static class SelvvirCommandTerminated extends RuntimeException {

		private static final long serialVersionUID = 1L;
	}

}
