package voicebot;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;
import java.util.Timer;
import java.util.TimerTask;

import org.jibble.pircbot.PircBot;
import org.jibble.pircbot.User;

import voicebot.AbstractVoiceBotCommand.SelvvirCommandTerminated;

// suggestions: statcounter browser use rates, would be good additions to the bot
public class VoiceBot extends PircBot {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String server = System.getProperty("server", "wolfe.freenode.net");
		String channel = System.getProperty("channel", "#pircbot");
		String name = System.getProperty("name", "GiskardReventlov");
		String password = System.getProperty("password", null);
		String verbose = System.getProperty("verbose", "");

		new VoiceBot(server, channel, name, password, "verbose".equals(verbose));
	}

	public static final String ALIAS_PREFIX = "alias:";

	protected static final String PREFIX = "`";

	private final String channel;
	private final String originalName;

	private final Map<String, AbstractVoiceBotCommand> commands;

	private final Map<String, Long> tracking;

	private final Map<String, String> preflight;

	private Timer aggressiveTakeBackTimer;
	private String accountPassword;

	private String lastMessage;
	private long lastMessageTime;

	public VoiceBot(String server, final String channel, String name, String password, boolean verbose) {
		commands = new LinkedHashMap<String, AbstractVoiceBotCommand>();

		ServiceLoader<AbstractVoiceBotCommand> loader = ServiceLoader.load(AbstractVoiceBotCommand.class);
		Iterator<AbstractVoiceBotCommand> itr = loader.iterator();

		while (itr.hasNext()) {
			AbstractVoiceBotCommand command = itr.next();
			commands.put(command.name(), command);
		}

		tracking = new LinkedHashMap<String, Long>();
		preflight = new HashMap<String, String>();
		this.channel = channel;

		try {
			setVerbose(verbose);
			setName(name);
			setLogin("giskard");
			setAutoNickChange(true);
			setEncoding("UTF-8");

			if (!server.isEmpty()) {

				if (password == null) {
					connect(server);
				}
				else {
					connect(server, 6667, password);
				}

				joinChannel(channel);
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		originalName = name;

		aggressiveTakeBackTimer = new Timer(true);
		accountPassword = password;

		final List<String> voiced = new ArrayList<String>();
		aggressiveTakeBackTimer.schedule(new TimerTask() {

			@Override
			public void run() {
				long date = new Date().getTime();

				synchronized (tracking) {
					for (String user : tracking.keySet()) {
						if (tracking.get(user) + (60000 * 10) <= date) {
							voice(channel, user);
							voiced.add(user);
						}
					}

					for (String user : voiced) {
						tracking.remove(user);
					}
					voiced.clear();
				}
			}
		}, 60000, 60000);
	}

	public Iterable<AbstractVoiceBotCommand> commands() {
		return commands.values();
	}

	public String getAliasTarget(String alias) {
		return alias.substring(ALIAS_PREFIX.length());
	}

	public void sendMessage(String channel, VoiceBotResponses response, String... values) {

		sendMessage(channel, String.format(response.toString(), (Object[]) values));
	}

	public void sendNotice(String channel, VoiceBotResponses response, String... values) {

		sendNotice(channel, String.format(response.toString(), (Object[]) values));
	}

	@Override
	protected void onConnect() {
		// Important! This prefixes all messages with + (for being logged in to an account) or - if not.
		sendRawLineViaQueue("CAP REQ identify-msg");
		sendRawLineViaQueue("CAP END");
	}

	@Override
	protected void onUserList(String channel, User[] users) {

		if (!originalName.equals(getNick())) {
			boolean found = false;

			for (User user : users) {
				if (originalName.equals(user.getNick())) {
					found = true;
					break;
				}
			}

			if (!found) {
				changeNick(originalName);
			}
		}
	}

	@Override
	protected void onMessage(String channel, String sender, String login, String hostname, String message) {

		handleMessage(channel, sender, login, hostname, message, false);
	}

	@Override
	protected void onPrivateMessage(String sender, String login, String hostname, String message) {

		if (message.equals("+``exit hyperspace") && ("AMcBain".equals(sender) || "Rainier".equals(sender))) {

			quitServer("bye!");
		}
		else if (sender.toLowerCase().equals("nickserv")) {
			String[] parts = message.split(" ");

			if (parts.length == 3 && parts[1].equals("ACC")) {
				if (parts[2].equals("3")) {
					// Handled by channel mode +q $~a
					//voice(preflight.get(parts[0]), parts[0]);
				}
				else {
					tracking.put(parts[0], new Date().getTime());
				}
				preflight.remove(parts[0]);
			}
		}
	}

	@Override
	protected void onJoin(String channel, String sender, String login, String hostname) {
		if (sender.equals(originalName) && channel.equals(this.channel)) {
			sendMessage("chanserv", "OP " + this.channel + " " + getNick());
		} else {
			sendMessage("nickserv", "ACC " + sender);
		}
	}

	private void handleMessage(String channel, String sender, String login, String hostname, String message, boolean privateMsg) {

		long now = new Date().getTime();
		boolean identified = "+".equals(message.substring(0, 1));
		message = message.substring(1);

		if (privateMsg && !message.startsWith(PREFIX)) {
			message = PREFIX + message;
		}

		if (message.startsWith(PREFIX)) {
			message = message.substring(PREFIX.length()).trim();

			if (message.isEmpty()) {
				return; // no reply
			}
			else if (message.equals(lastMessage) && now - lastMessageTime <= 15000) {
				return; // anti-spam
			}

			lastMessage = message;
			lastMessageTime = now;

			String target = sender;

			if (message.contains(" @ ")) {
				int at = message.indexOf(" @ ");
				String after = message.substring(at + 3);
				message = message.substring(0, at).trim();

				if (!after.isEmpty()) {
					target = after.trim();
				}
			}

			String command = message;
			String flags = "";
			int space = message.indexOf(' ');

			if (space != -1) {
				command = message.substring(0, space);
			}

			int slash = command.indexOf('/');
			if (slash != -1) {

				flags = command.substring(slash + 1);
				command = command.substring(0, slash);
			}

			AbstractVoiceBotCommand cmd = commands.get(command);
			if (cmd != null) {
				message = (space == -1 ? "" : message.substring(space + 1)).trim();

				if (!cmd.isAdminOnly() || identified) {

					runCommand(cmd, flags, channel, sender, target, message);
				}
				else {

					sendNotice(sender, VoiceBotResponses.ADMIN_ONLY);
				}
			}
			else {
				sendNotice(sender, VoiceBotResponses.NOT_FOUND, message);
			}
		}
	}

	private void runCommand(AbstractVoiceBotCommand cmd, String flags, String channel, String sender, String target, String message) {

		try {
			cmd.run(this, flags, channel, sender, target, message);
		}
		catch (SelvvirCommandTerminated t) {
			// this is their way of exiting on error and not processing anything more
			// makes their code cleaner, though this gives me a feeling of exception-based control/flow :(
		}
	}

	@Override
	protected void onQuit(String sourceNick, String sourceLogin, String sourceHostname, String reason) {

		if (originalName.equals(sourceNick)) {
			changeNick(originalName);
		}

		synchronized (tracking) {
			if (tracking.containsKey(sourceNick)) {
				tracking.remove(sourceNick);
			}
		}
	}

	@Override
	protected void onNickChange(String oldNick, String login, String hostname, String newNick) {

		if (originalName.equals(oldNick) && accountPassword != null) {

			aggressiveTakeBackTimer.schedule(new TimerTask() {

				private int sequence;

				@Override
				public void run() {
					switch (sequence) {
					case 0:
						sendMessage("NickServ", "identify " + originalName + " " + accountPassword);
						break;
					case 1:
						sendMessage("NickServ", "release " + originalName);
						break;
					case 2:
						changeNick(originalName);
						cancel();
					}
					sequence++;
				}
			}, 3000, 3000);
		}
	}

	@Override
	protected void onDisconnect() {
		dispose();
	}
}
