package controllers;

import play.*;
import play.mvc.*;

import java.util.*;

import models.*;

import java.lang.reflect.*;
import selvvir.*;
import java.util.prefs.Preferences;

import com.zwitserloot.json.JSON;

public class Application extends Controller {

    private static final ServiceLoader<AbstractSelvvirCommand> LOADER;
    private static Preferences PREFS;
    private static Preferences STATS;
    static {
        LOADER = ServiceLoader.load(AbstractSelvvirCommand.class);
        try {
            Field key = Selvvir.class.getDeclaredField("FACTOIDS_KEY");
            key.setAccessible(true);
            PREFS = Preferences.userNodeForPackage(Selvvir.class).node((String)key.get(null) + "-libera");
            key = Selvvir.class.getDeclaredField("STATISTICS_KEY");
            key.setAccessible(true);
            STATS = Preferences.userNodeForPackage(Selvvir.class).node((String)key.get(null) + "-libera");
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public static void index() {
        ServiceLoader<?> navigation = LOADER;
        ServiceLoader<?> commands = LOADER;
        render(navigation, commands);
    }

    public static void factoids() {
        List<String> keys = Collections.emptyList();
        try {
            PREFS.sync();
            keys = Arrays.asList(PREFS.keys());
        } catch(Exception e) {
            // FUCKING JAVA!
            e.printStackTrace();
        }
        Preferences values = PREFS;
        render(keys, values);
    }

    public static void apiFactoidsSimple() {
        StringBuilder keys = new StringBuilder("{");
        try {
            PREFS.sync();
            String[] prefs = PREFS.keys();
            Arrays.sort(prefs);

            for (int i = 0; i < prefs.length; i++) {
                String pref = prefs[i];
                keys.append('"');
                keys.append(pref.replace("\"", "\\\""));
                keys.append("\":");

                String value = PREFS.get(pref, "");
                keys.append('"');
                keys.append(value.replace("\"", "\\\""));
                keys.append('"');

                if (i < prefs.length - 1) {
                    keys.append(",");
                }
            }
        } catch(Exception e) {
            // FUCKING JAVA!
            e.printStackTrace();
        }
        keys.append("}");
        renderJSON(keys.toString());
    }

    public static void apiFactoids(boolean flat) {
        StringBuilder keys = new StringBuilder("[");
        try {
            PREFS.sync();
            String[] prefs = PREFS.keys();
            Arrays.sort(prefs);

            for (int i = 0; i < prefs.length; i++) {
                String pref = prefs[i];
                keys.append("{\"name\":\"");
                keys.append(pref.replace("\"", "\\\""));
                keys.append('"');

                String value = PREFS.get(pref, "");
                if (value.startsWith("alias:")) {
                    if (flat) {
                        value = PREFS.get(value.split(":")[1], "");
                    } else {
                        keys.append(",\"alias\":true");
                    }
                }
                keys.append(",\"value\":\"");
                keys.append((value.startsWith("alias:")? value.split(":")[1] : value).replace("\"", "\\\""));
                keys.append('"');

                int j = 0;
                boolean flag = false;
                for (String bit : SelvvirJavaExtensions.urlify(value)) {
                    if (bit.startsWith("+")) {
                        if (!flag) {
                            keys.append(",\"urls\":[");
                            flag = true;
                        }
                        keys.append('[');
                        keys.append(j);
                        keys.append(',');
                        keys.append(j + bit.length() - 1);
                        keys.append(']');
                        keys.append(',');
                    }
                    j += bit.length() - 1;
                }
                if (flag) {
                    keys.deleteCharAt(keys.length() - 1);
                }
                keys.append(flag? "],": ',');

                int hits = STATS.getInt(pref, 0);
                keys.append("\"hits\":");
                keys.append(hits);
                keys.append('}');

                if (i < prefs.length - 1) {
                    keys.append(",");
                }
            }
        } catch(Exception e) {
            // FUCKING JAVA!
            e.printStackTrace();
        }
        keys.append("]");
        renderJSON(keys.toString());
    }

    public static void apiFactoidUsages() {
        JSON json = JSON.newMap();
        try {
            PREFS.sync();
            STATS.sync();
            String[] prefs = PREFS.keys();

            for (int i = 0; i < prefs.length; i++) {
                String pref = prefs[i];

                int value = STATS.getInt(pref, 0);
                JSON obj = json.get(pref);
                if (!obj.exists())
                {
                    obj.get("total").setInt(value);
                }

                String target = PREFS.get(pref, "");
                if (target.startsWith("alias:")) {

                    target = target.split(":")[1];
                    JSON t = json.get(target);
                    if (!t.exists())
                    {
                        int other = STATS.getInt(target, 0);
                        t.get("total").setInt(other);
                    }
                    if (!t.get("own").exists()) {
                        t.get("own").setInt(t.get("total").asInt());
                    }
                    t.get("own").setInt(t.get("own").asInt() - value);
                }
            }
        } catch(Exception e) {
            // WARGH.
            e.printStackTrace();
        }
        renderJSON(json.toJSON());
    }

    public static void apiCommands() {
        StringBuilder output = new StringBuilder();
        output.append("{");

        try {
            for (AbstractSelvvirCommand command : LOADER) {
                output.append('"');
                output.append(command.name().replace("\"", "\\\""));
                output.append("\":{\"syntax\":\"");
                output.append(SelvvirJavaExtensions.syntax(command).replace("\"", "\\\"").trim());
                output.append("\",\"description\":\"");
                output.append(SelvvirJavaExtensions.description(command).replace("\"", "\\\""));
                output.append("\"");
                if (command.isAdminOnly()) {
                    output.append(",\"admin\":true");
                }
                if (SelvvirJavaExtensions.hasFlags(command)) {
                    output.append(",\"flags\":{");
                    for (Map.Entry<String, String> flag : SelvvirJavaExtensions.flags(command, "")) {
                        output.append('"');
                        output.append(flag.getKey().replace("\"", "\\\""));
                        output.append("\":\"");
                        output.append(flag.getValue().replace("\"", "\\\""));
                        output.append("\",");
                    }
                    output.deleteCharAt(output.length() - 1);
                    output.append("}");
                }
                if (SelvvirJavaExtensions.hasNotes(command)) {
                    output.append(",\"notes\":[");
                    for (String note : SelvvirJavaExtensions.notes(command)) {
                        output.append('"');
                        output.append(note.replace("\"", "\\\""));
                        output.append("\",");
                    }
                    output.deleteCharAt(output.length() - 1);
                    output.append("]");
                }
                output.append("},");
            }
            output.deleteCharAt(output.length() - 1);
        } catch(Exception e) {
            // BLARGH.
            e.printStackTrace();
        }

        output.append("}");
        renderJSON(output.toString());
    }
}
