package controllers;

import play.templates.JavaExtensions;
import selvvir.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.regex.*;

public class SelvvirJavaExtensions extends JavaExtensions {

    private static final Pattern URL_REGEX = Pattern.compile("(?i)\\b((?:[a-z][\\w-]+:(?:/{1,3}|[a-z0-9%])|www\\d{0,3}[.]|[a-z0-9.\\-]+[.][a-z]{2,4}/)(?:[^\\s()<>]+|\\(([^\\s()<>]+|(\\([^\\s()<>]+\\)))*\\))+(?:\\(([^\\s()<>]+|(\\([^\\s()<>]+\\)))*\\)|[^\\s`!()\\[\\]{};:'\".,<>?«»“”‘’]))");

    private static Field ARGS;
    private static Field INFO;
    private static Field HELP;
    private static Field KEYS;
    private static Field NOTES;

    static {
        try {
            ARGS = AbstractSelvvirCommand.class.getDeclaredField("args");
            ARGS.setAccessible(true);
            INFO = AbstractSelvvirCommand.class.getDeclaredField("info");
            INFO.setAccessible(true);
            HELP = AbstractSelvvirCommand.class.getDeclaredField("help");
            HELP.setAccessible(true);
            KEYS = AbstractSelvvirCommand.class.getDeclaredField("keys");
            KEYS.setAccessible(true);
            NOTES = AbstractSelvvirCommand.class.getDeclaredField("notes");
            NOTES.setAccessible(true);
        } catch(NoSuchFieldException e) {
            e.printStackTrace();
        }
    }

    public static String syntax(AbstractSelvvirCommand command) throws IllegalAccessException {
        String args = (String)ARGS.get(command);
        return command.name() + " " + args;
    }

    public static String description(AbstractSelvvirCommand command) throws IllegalAccessException {
        return (String)INFO.get(command);
    }

    public static boolean hasFlags(AbstractSelvvirCommand command) throws IllegalAccessException {
        return !((Map<String, String>)HELP.get(command)).isEmpty();
    }

    public static Set<Map.Entry<String, String>> flags(AbstractSelvvirCommand command, String flag) throws IllegalAccessException {
        return ((Map<String, String>)HELP.get(command)).entrySet();
    }

    public static boolean hasNotes(AbstractSelvvirCommand command) throws IllegalAccessException {
        return ((String[])NOTES.get(command)).length > 0;
    }

    public static List<String> notes(AbstractSelvvirCommand command) throws IllegalAccessException {
        return Arrays.asList((String[])NOTES.get(command));
    }

    public static List<String> urlify(String input) {
        Matcher matches = URL_REGEX.matcher(input);
        List<String> list = new ArrayList<String>();

        int prev = 0;
        while(matches.find()) {

            list.add("-" + input.substring(prev, matches.start()));
            list.add("+" + matches.group());
            prev = matches.end();
        }
        list.add("-" + input.substring(prev));

        return list;
    }
}
